#define WIN32_LEAN_AND_MEAN
#include <Fl/x.H>
#include "Fl_Icon_Window.h"
#include <Fl/Fl_Image.H>
#include <Fl/fl_draw.h>

#include <stdlib.h>
#include <stdint.h>
#include <string.h>

// The monochrome bitmask is set to transparent if the Alpha channel is less than this.
// It probably should not be zero.
#define BITMASK_THRESHOLD 0x20

// TODO:
//
// Put the "init" stuff into FLTK's initialization.
//
// Figure out how to handle Large and Small icons:
//
//   Maybe just accumulate all defined icons that don't duplicate a size,
//   and allow the set to be flushed by ::icon(NULL) ?
//
//   Mac OSX and Win32 both have 16x16 and 32x32 as the standard small/large sizes,
//   and accept additional sizes as well (usually larger). In X11, Pixmap icons are one
//   size only. NET_WM icons are multiple sizes.
//
//   For now, it should work OK for all systems to provide just the 32x32 icon.
//
// The X11 code checks for NET_WM_ICON support, and uses it if available.
//
// There is no Mac code included. Generating run-time icons is a bit difficult there.
// OSX may also have mandatory icon size rules, so these would may need to be derived.

static void init_icon_support (void);

Fl_Icon_Window::Fl_Icon_Window (int W, int H, const char *l):FL_WINDOW_ (W, H, l)
{
  init_icon_support ();
}

Fl_Icon_Window::Fl_Icon_Window (int X, int Y, int W, int H,
                                const char *l):FL_WINDOW_ (X, Y, W, H, l)
{
  init_icon_support ();
}


/////////////////////////////////////////////////////////////////////////////////////
#ifdef WIN32

static SIZE small_icon = { 0, 0 }, large_icon = {
0, 0};

// Windows has a different large/small set of sizes for SHELL icons:
// SHGetFileInfo function with SHGFI_SHELLICONSIZE | SHGFI_SMALLICON

static void init_icon_support (void)
{
  if (!small_icon.cx) {
    small_icon.cx = GetSystemMetrics (SM_CXICON);
    small_icon.cy = GetSystemMetrics (SM_CYICON);
    large_icon.cx = GetSystemMetrics (SM_CXSMICON);
    large_icon.cy = GetSystemMetrics (SM_CYSMICON);
  }
}

void Fl_Icon_Window::icon (const Fl_Image * image)
{
  if (!image)
    return;
  HICON hIcon = ImageToIcon (image);

  Fl_Window::icon ((char *) hIcon);
  if (shown ())
    SendMessage (fl_xid (this), WM_SETICON, ICON_BIG, (LPARAM) hIcon);
};

// NOTES:
// CreateIcon and CreateIconIndirect traditionally require DDBitmaps.
// WinXP accepts 32-bit BGRA DIBitmaps with full 8-bit alpha support.
// Strange but true for Compatible Bitmaps:
//   If made from the Screen DC, the format matches the screen.
//   If made from a "Screen Compatible" DC, it is monochrome.
//   In other words:
//   CreateCompatibleBitmap(GetDC(NULL),...) == COLOR
//   CreateCompatibleBitmap(CreateCompatibleDC(GetDC(NULL)),...) == MONO
//
// The icon is created as follows:
// 1) Draw the image to a 32-bit BGRA DIB.
// 2) Create a monochrome mask bitmap. (This is done using SetPixelV during
//    step 2; easier but less efficient than raw bit manipulation.)
// 3) Try to create the Icon using the DIB+Mask.
// 4) If that fails, reduce the Icon to the standard DDB+Mask and try again.

HICON Fl_Icon_Window::ImageToIcon (const Fl_Image * image)
{
  typedef struct _BGRA_PIXEL {
    unsigned char b, g, r, a;
  } BGRA_PIXEL;

  int width = image->w ();
  int height = image->h ();
  int pixelstride = image->d ();
  int rowstride = image->ld ();
  int depth = abs (pixelstride);
  if (depth>4) depth=4;

  if (!rowstride)
    rowstride = width * pixelstride;

  BITMAPINFO bmi = { {sizeof (BITMAPINFOHEADER), width, height, 1, depth*8, BI_RGB } };
  bmi.bmiHeader.biBitCount = 32;

  // All bitmaps have 16-bit aligned rows, some have 32-bit aligned rows.
  // These DIBs should all be 32-bit aligned, but I use GetObject here to ensure
  // compatibility with OS variants.
  HDC dcScreen = GetDC (NULL);
  BYTE *dibColorData;
  HBITMAP dibColor = CreateDIBSection (dcScreen, &bmi,
                  DIB_RGB_COLORS, (void **)&dibColorData, NULL, 0);
  BITMAP dibColorDesc;
  GetObject (dibColor, sizeof (BITMAP), &dibColorDesc);

  HDC dcMask = CreateCompatibleDC (dcScreen);
  HBITMAP bmpMask = CreateCompatibleBitmap (dcMask, width, height);
  HGDIOBJ hgdiMaskDefault = SelectObject (dcMask, bmpMask);

  const BYTE *image_data = (const BYTE *) *image->data ();
  int dib_rowstride = dibColorDesc.bmWidthBytes >> 2;
  for (int y = 0; y < height; y++) {
    int bmp_y = height - y - 1;
    const BYTE *image_pixel = image_data + y * rowstride;
    BGRA_PIXEL *outrow = (BGRA_PIXEL *) dibColorDesc.bmBits + bmp_y * dib_rowstride;
    for (int x = 0; x < width; x++) {
      switch (depth) {
      case 1:
        outrow[x].r = outrow[x].g = outrow[x].b = *image_pixel;
        outrow[x].a = 0xff;
        SetPixelV(dcMask,x,y,RGB(0,0,0));
        break;
      case 3:
        outrow[x].r = image_pixel[0];
        outrow[x].g = image_pixel[1];
        outrow[x].b = image_pixel[2];
        outrow[x].a = 0xff;
        SetPixelV(dcMask,x,y,RGB(0,0,0));
        break;
      case 4:
        int a = outrow[x].a = image_pixel[3];
        if (a > BITMASK_THRESHOLD) SetPixelV(dcMask,x,y,RGB(0,0,0));
        else SetPixelV(dcMask,x,y,RGB(255,255,255));
        // Windoze Bitmaps must be pre-multiplied by Alpha:
        outrow[x].r = (a * (int)image_pixel[0]) / 255;
        outrow[x].g = (a * (int)image_pixel[1]) / 255;
        outrow[x].b = (a * (int)image_pixel[2]) / 255;
        break;
      }
      image_pixel += pixelstride;
    }
  }
  SelectObject(dcMask, hgdiMaskDefault);
  DeleteDC(dcMask);

  HICON icon = NULL;

  // This probably only works in WinXP; otherwise a DDB is required.
  if (depth == 4) {
    ICONINFO ii_bgra = { TRUE, 0, 0, bmpMask, dibColor };
    icon = CreateIconIndirect (&ii_bgra);
  }

  if (!icon) {
    HDC dcColor = CreateCompatibleDC (dcScreen);
    HBITMAP ddbColor = CreateCompatibleBitmap (dcScreen, width, height);
    HGDIOBJ hgdiColorDefault = SelectObject (dcColor, ddbColor);
    SetDIBits (dcColor, ddbColor, 0, height, dibColorData, &bmi,
               DIB_RGB_COLORS);
    SelectObject (dcColor, hgdiColorDefault);
    DeleteDC (dcColor);

    ICONINFO ii_ddb = { TRUE, 0, 0, bmpMask, ddbColor };
    icon = CreateIconIndirect (&ii_ddb);
    DeleteObject (ddbColor);
  }

  ReleaseDC (NULL, dcScreen);
  DeleteObject (dibColor);
  DeleteObject (bmpMask);

  return icon;
}

////////////////////////////////////////////////////////////////////////
#elif defined(__APPLE__)
# error No Mac support yet.

////////////////////////////////////////////////////////////////////////
#else // Assume POSIX/X11

static int net_wm_icon_supported = 0;

static Atom *net_supported = NULL;
static long int net_supported_nitems = 0;
static XIconSize *icon_sizes = NULL;
static int icon_size_count=0;

static void init_icon_support (void)
{
  static int initialized=0;
  if (!initialized) {
    initialized=1;

    if (!fl_display) fl_open_display ();
    Atom _NET_SUPPORTED = XInternAtom (fl_display, "_NET_SUPPORTED", True /* only_if_exists */);
    if (_NET_SUPPORTED) {
      Atom type;
      unsigned long int bytes_after;
      int format;
      XGetWindowProperty (fl_display, RootWindow (fl_display, fl_screen),
                        _NET_SUPPORTED, 0, 65536 / sizeof (int32_t), False, XA_ATOM,
                        &type, &format, (unsigned long int *) &net_supported_nitems,
                        &bytes_after, (unsigned char **) &net_supported);
    }
    XGetIconSizes(fl_display, RootWindow (fl_display, fl_screen),&icon_sizes,&icon_size_count);
  }
}

Pixmap Fl_Icon_Window::ImageToPixmap (const Fl_Image * image)
{
  Pixmap pixmap =
    XCreatePixmap (fl_display, RootWindow (fl_display, fl_screen), image->w (),
                   image->h (), fl_visual->depth);
  // Use fl_draw_image() to draw to the Pixmap. If there is no current Window, we must
  // also provide a GC. (O
  if (pixmap) {
    GC current_gc = fl_gc;
    Window current_window = fl_window;
    fl_window = pixmap;
    if (!fl_gc) fl_gc = XCreateGC(fl_display, pixmap, 0, 0);
    fl_draw_image ((const uchar *) *image->data (), 0, 0, image->w (), image->h (), image->d ());
    if (!current_gc) XFreeGC(fl_display,fl_gc);
    fl_window = current_window;
  }
  return pixmap;
}

Pixmap Fl_Icon_Window::ImageToBitmap (const Fl_Image * image)
{
  int pixel_delta = image->d ();
  if (pixel_delta < 4 && pixel_delta > -4)
    return None;
  int line_delta = image->ld ();
  int w = image->w ();
  int h = image->h ();
  if (!line_delta)
    line_delta = pixel_delta * w;

  int line_size = (w + 7) >> 3; // bytes-per-line in bitmap data
  char *mask_data = new char[line_size * h];
  memset (mask_data, 0, line_size * h);
  uint8_t *pixels = (uint8_t *) * image->data ();
  uint8_t *p, *row;
  int x, y;
  row = pixels;
  char *mask_row = mask_data;
  for (y = 0, row = pixels; y < h; y++, row += line_delta) {
    for (x = 0, p = row; x < w; x++, p += pixel_delta) {
      if (p[3] > BITMASK_THRESHOLD)
        mask_row[x >> 3] |= (1 << (x & 7));
    }
    mask_row += line_size;
  }
  // This creates a 1-bit mask from raw binary data, with 8-bit aligned rows.
  Pixmap pixmap = XCreateBitmapFromData (fl_display,
                                         RootWindow (fl_display, fl_screen),
                                         mask_data, w, h);
  delete[]mask_data;
  return pixmap;
}

uint32_t *Fl_Icon_Window::ImageToNetIcon (const Fl_Image * image)
{
  int pixel_delta = image->d ();
  int line_delta = image->ld ();
  int w = image->w ();
  int h = image->h ();
  if (!line_delta) line_delta = pixel_delta * w;
  uint8_t *pixels = (uint8_t *) * image->data ();
  uint32_t *net_icon = new uint32_t[w * h + 2];
  uint8_t *p, *row;
  uint32_t argb, *icon_pixel = &net_icon[2];
  int x, y;
  net_icon[0] = w;
  net_icon[1] = h;
  row = pixels;
  for (y = 0, row = pixels; y < h; y++, row += line_delta) {
    for (x = 0, p = row; x < w; x++, p += pixel_delta) {
      switch (abs (pixel_delta)) {
      case 1:
        argb = ((uint32_t) p[0] << 16) | ((uint32_t) p[0] << 8) | (uint32_t) p[0]
          | 0xff000000;
        break;
      case 3:
        argb = ((uint32_t) p[0] << 16) | ((uint32_t) p[1] << 8) | (uint32_t) p[2]
          | 0xff000000;
        break;
      case 4:
        argb = ((uint32_t) p[0] << 16) | ((uint32_t) p[1] << 8) | (uint32_t) p[2]
          | ((uint32_t) p[3] << 24);
        break;
      }
      *icon_pixel++ = argb;
    }
  }
  return net_icon;
}

void Fl_Icon_Window::icon (const Fl_Image * image)
{
  if (!image)
    return;
  if (net_wm_icon_supported) {
    uint32_t *net_icon = ImageToNetIcon(image);
    Fl_Window::icon ((char *) net_icon);
    if (shown ()) {
      XChangeProperty (fl_display, fl_xid (this),
                       XInternAtom (fl_display, "_NET_WM_ICON", 0), XA_CARDINAL, 32,
                       PropModeReplace, (unsigned char *) net_icon,
                       net_icon[0] * net_icon[1] + 2);
    }
  } else {
    Pixmap pixmap = ImageToPixmap(image);
    Fl_Window::icon ((char *) pixmap);
    if (shown ()) {
      Pixmap mask = ImageToBitmap(image);
      XWMHints *hints = XAllocWMHints ();
      hints->flags = IconPixmapHint | IconMaskHint;
      hints->icon_pixmap = pixmap;
      hints->icon_mask = mask;
      XSetWMHints (fl_display, fl_xid (this), hints);
      XFree (hints);
    }
  }
}
#endif
